# =============================================================================
# LLM Provider Configuration
# =============================================================================
{% if primary_provider == "openrouter" %}
# OpenRouter API Key (https://openrouter.ai/keys)
OPENROUTER_API_KEY=sk-or-v1-...
{% elif primary_provider == "anthropic" %}
# Anthropic API Key (https://console.anthropic.com/settings/keys)
ANTHROPIC_API_KEY=sk-ant-api03-...
{% elif primary_provider == "azure" %}
# Azure OpenAI Configuration
AZURE_API_KEY=...
AZURE_API_BASE=https://your-resource.openai.azure.com/
AZURE_API_VERSION=2024-02-15-preview
{% elif primary_provider == "google" or primary_provider == "gemini" %}
# Google Gemini API Key (https://aistudio.google.com/app/apikey)
GEMINI_API_KEY=...
{% elif primary_provider == "bedrock" %}
# AWS Bedrock Configuration
AWS_ACCESS_KEY_ID=...
AWS_SECRET_ACCESS_KEY=...
AWS_REGION_NAME=us-east-1
{% else %}
# OpenAI API Key (https://platform.openai.com/api-keys)
OPENAI_API_KEY=sk-proj-...
{% endif %}

# LLM Model (configured from spec: {{ primary_model }})
LLM_MODEL={{ primary_model }}
# SUMMARIZER_MODEL=
# REFLECTION_MODEL=

# =============================================================================
# Feature Flags
# =============================================================================
MEMORY_ENABLED={{ memory_enabled }}
SUMMARIZER_ENABLED={{ summarizer_enabled }}
REFLECTION_ENABLED={{ reflection_enabled }}

# =============================================================================
# Reflection Settings (if enabled)
# =============================================================================
# REFLECTION_QUALITY_THRESHOLD=0.8
# REFLECTION_MAX_REVISIONS=2

# =============================================================================
# Service URLs
# =============================================================================
MEMORY_BASE_URL={{ memory_base_url|default("http://localhost:8000", true) }}
LIGHTHOUSE_BASE_URL={{ lighthouse_base_url|default("http://localhost:8081", true) }}
WAYFINDER_BASE_URL={{ wayfinder_base_url|default("http://localhost:8082", true) }}

# =============================================================================
# Planner Settings
# =============================================================================
PLANNER_MAX_ITERS={{ planner_max_iters }}
PLANNER_HOP_BUDGET={{ planner_hop_budget }}
PLANNER_ABSOLUTE_MAX_PARALLEL={{ planner_absolute_max_parallel }}
PLANNER_STREAM_FINAL_RESPONSE={{ planner_stream_final_response }}
